;; Generated package description from fontaine.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "fontaine" "3.0.1.0.20260316.175802" "Set font configurations using presets" '((emacs "29.1")) :commit "3c1401de103af1dd38dc50387ceab7b0fd81b15d" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Protesilaos Stavrou" . "info@protesilaos.com") :url "https://github.com/protesilaos/fontaine")
